﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace Inedo
{
    /// <summary>
    /// Contains All-purpose Helper methods
    /// </summary>
    public static partial class AH
    {
        private const int SecondsPerMinute = 60;
        private const int SecondsPerHour = 60 * SecondsPerMinute;
        private const int SecondsPerDay = 24 * SecondsPerHour;
        private const int SecondsPerMonth = 30 * SecondsPerDay;

        /// <summary>
        /// Converts the specified string representation of a date and time to its <see cref="DateTime"/> equivalent
        /// using <see cref="DateTime.TryParse(string, out System.DateTime)"/>; if the string is not a valid date time, null is returned
        /// </summary>
        /// <param name="dateTimeString">String to try to parse.</param>
        /// <returns>Resulting <see cref="System.DateTime"/> value if parsing was successful; otherwise null.</returns>
        public static DateTime? ParseDate(string dateTimeString)
        {
            DateTime dt;
            if (DateTime.TryParse(dateTimeString, out dt))
                return dt;
            else
                return null;
        }

        /// <summary>
        /// Calculates the relative time from the specified dates
        /// </summary>
        /// <param name="referenceFromDate">The reference <see cref="System.DateTime"/> used to calculate relative time. Typically, this is <see cref="System.DateTime.Now"/>.</param>
        /// <param name="toDate">The <see cref="System.DateTime"/> with which the reference value will be compared.</param>
        /// <returns><see cref="System.String"/> containing relative time if dates are valid, otherwise an empty string.</returns>
        public static string RelativeDate(DateTime? referenceFromDate, DateTime? toDate)
        {
            if (!referenceFromDate.HasValue || !toDate.HasValue)
                return string.Empty;

            bool useFutureSuffix = (toDate > referenceFromDate);

            var time = (useFutureSuffix)
                ? toDate.Value.Subtract(referenceFromDate.Value)
                : referenceFromDate.Value.Subtract(toDate.Value);

            string suffix = (useFutureSuffix) ? "from now" : "ago";

            if (time.TotalSeconds < 24 * SecondsPerHour)
            {
                return "today";
            }
            if (time.TotalSeconds < 48 * SecondsPerHour)
            {
                return useFutureSuffix ? "tomorrow" : "yesterday";
            }
            if (time.TotalSeconds < 30 * SecondsPerDay)
            {
                return time.Days + " days " + suffix;
            }
            if (time.TotalSeconds < 12 * SecondsPerMonth)
            {
                int months = Convert.ToInt32(Math.Floor((double)time.Days / 30));
                return months <= 1 ? "one month " + suffix : months + " months " + suffix;
            }
            else
            {
                int years = Convert.ToInt32(Math.Floor((double)time.Days / 365));
                return years <= 1 ? "one year " + suffix : years + " years " + suffix;
            }
        }

        /// <summary>
        /// Converts the specified string representation of an integer to its <see cref="int"/> equivalent
        /// using <see cref="int.TryParse(string, out int)"/>. If the string is not a valid <see cref="int"/>, <c>null</c>
        /// is returned.
        /// </summary>
        /// <param name="intString">String to try to parse.</param>
        /// <returns>Resulting <see cref="int"/> value if parsing was successful; otherwise null.</returns>
        public static int? ParseInt(string intString)
        {
            int i;
            if (int.TryParse(intString, out i))
                return i;
            else
                return null;
        }

        internal static class UserAccounts
        {
            public static void CanonicalizeAccountName(ref string userAccount)
            {
                if (string.IsNullOrEmpty(userAccount))
                    return;

                var match = Regex.Match(userAccount, @"^((?<1>[^\\]*)\\)?(?<2>.+)$", RegexOptions.ExplicitCapture);
                var name = match.Groups[2].Value;

                if (string.Equals(name, "LocalSystem", StringComparison.OrdinalIgnoreCase) || string.Equals(name, "LOCAL SYSTEM", StringComparison.OrdinalIgnoreCase))
                    userAccount = "NT AUTHORITY\\LocalSystem";
                else if (string.Equals(name, "NetworkService", StringComparison.OrdinalIgnoreCase) || string.Equals(name, "NETWORK SERVICE", StringComparison.OrdinalIgnoreCase))
                    userAccount = "NT AUTHORITY\\NetworkService";
            }
        }

        /// <summary>
        /// Returns a null reference if the two specified parameters are equal (or if either are null or DBNull.Value); otherwise, returns the first parameter.
        /// </summary>
        /// <param name="a">The first parameter.</param>
        /// <param name="b">The second parameter.</param>
        /// <returns>Null reference if the two specified parameters are equal (or if either are null or DBNull.Value); otherwise, the first parameter.</returns>
        public static object NullIf(object a, object b)
        {
            if (a == null || Convert.IsDBNull(a) || b == null || Convert.IsDBNull(b) || a.Equals(b))
                return null;

            return a;
        }

        /// <summary>
        /// Returns a null reference if the two specified parameters are equal (or if either are null); otherwise, returns the first parameter.
        /// </summary>
        /// <param name="a">The first parameter.</param>
        /// <param name="b">The second parameter.</param>
        /// <returns>Null reference if the two specified parameters are equal (or if either are null); otherwise, the first parameter.</returns>
        public static string NullIf(string a, string b)
        {
            return NullIf((object)a, (object)b) as string;
        }

        /// <summary>
        /// Returns a null reference if the two specified parameters are equal (or if either are null); otherwise, returns the first parameter.
        /// </summary>
        /// <param name="a">The first parameter.</param>
        /// <param name="b">The second parameter.</param>
        /// <typeparam name="T">The type of parameter to compare.</typeparam>
        /// <returns>Null reference if the two specified parameters are equal (or if either are null); otherwise, the first parameter.</returns>
        public static T? NullIf<T>(T? a, T? b)
            where T : struct
        {
            if (a == null || b == null || a.Equals(b))
                return null;
            else
                return a;
        }
        /// <summary>
        /// Returns a null reference if the two specified parameters are equal (or if either are null); otherwise, returns the first parameter.
        /// </summary>
        /// <param name="a">The first parameter.</param>
        /// <param name="b">The second parameter.</param>
        /// <typeparam name="T">The type of parameter to compare.</typeparam>
        /// <returns>Null reference if the two specified parameters are equal (or if either are null); otherwise, the first parameter.</returns>
        public static T? NullIf<T>(T a, T? b)
            where T : struct
        {
            if (b == null || a.Equals(b))
                return null;
            else
                return a;
        }

        /// <summary>
        /// Returns the first non-null and non-<see cref="string.Empty"/> expression among its arguments.
        /// </summary>
        /// <param name="args">Objects to convert to strings and test for null or empty.</param>
        /// <returns>First non-null and non-<see cref="string.Empty"/> value.</returns>
        public static string CoalesceStr(params string[] args)
        {
            foreach (object arg in args)
            {
                var s = arg?.ToString();
                if (!string.IsNullOrEmpty(s))
                    return s;
            }

            return null;
        }

        /// <summary>
        /// Returns <see cref="string.Empty"/> when one of the arguments is null, <see cref="DBNull.Value"/>, or <see cref="string.Empty"/>; 
        /// otherwise, returns a concatenation of the arguments.
        /// </summary>
        /// <param name="args">Arguments to concatenate.</param>
        /// <returns>Concatenated string or <see cref="string.Empty"/> if an argument is <c>null</c>, 
        /// <see cref="DBNull.Value"/>, or <see cref="string.Empty"/>.</returns>
        public static string ConcatNE(params object[] args)
        {
            if (args == null)
                throw new ArgumentNullException(nameof(args));

            if (args.Any(a => a == null || string.Empty.Equals(a) || Convert.IsDBNull(a)))
                return string.Empty;
            else
                return string.Concat(args);
        }

        /// <summary>
        /// Begins a switch expression.
        /// </summary>
        /// <typeparam name="TArg">The type of the input argument.</typeparam>
        /// <typeparam name="TResult">The type of the result.</typeparam>
        /// <param name="input">The input value.</param>
        /// <returns>The switch expression.</returns>
        [Undisclosed]
        public static IDefaultSwitch<TArg, TResult> Switch<TArg, TResult>(TArg input)
        {
            return Switch<TArg, TResult>(input, null);
        }
        /// <summary>
        /// Begins a switch expression.
        /// </summary>
        /// <typeparam name="TArg">The type of the input argument.</typeparam>
        /// <typeparam name="TResult">The type of the result.</typeparam>
        /// <param name="input">The input value.</param>
        /// <param name="comparer">Equality comparer to use for the input value.</param>
        /// <returns>The switch expression.</returns>
        [Undisclosed]
        public static IDefaultSwitch<TArg, TResult> Switch<TArg, TResult>(TArg input, IEqualityComparer<TArg> comparer)
        {
            return new SwitchBlock<TArg, TResult>(input, comparer ?? EqualityComparer<TArg>.Default);
        }

        private sealed class SwitchBlock<TArg, TResult> : IDefaultSwitch<TArg, TResult>, ISwitch<TArg, TResult>
        {
            private TArg input;
            private IEqualityComparer<TArg> comparer;
            private Func<TResult> getDefaultValue;
            private Func<TResult> getResult;

            public SwitchBlock(TArg input, IEqualityComparer<TArg> comparer)
            {
                this.input = input;
                this.comparer = comparer;
            }

            IDefaultSwitch<TArg, TResult> IDefaultSwitch<TArg, TResult>.Case(TArg value, TResult result)
            {
                return this.Case(value, () => result);
            }
            IDefaultSwitch<TArg, TResult> IDefaultSwitch<TArg, TResult>.Case(TArg value, Func<TResult> getResult)
            {
                return this.Case(value, getResult);
            }

            ISwitch<TArg, TResult> IDefaultSwitch<TArg, TResult>.Default(TResult result)
            {
                this.getDefaultValue = () => result;
                return this;
            }
            ISwitch<TArg, TResult> IDefaultSwitch<TArg, TResult>.Default(Func<TResult> getResult)
            {
                if (getResult == null)
                    throw new ArgumentNullException("getResult");

                this.getDefaultValue = getResult;
                return this;
            }

            ISwitch<TArg, TResult> ISwitch<TArg, TResult>.Case(TArg value, TResult result)
            {
                return this.Case(value, () => result);
            }
            ISwitch<TArg, TResult> ISwitch<TArg, TResult>.Case(TArg value, Func<TResult> getResult)
            {
                return this.Case(value, getResult);
            }

            TResult ISwitch<TResult>.End()
            {
                if (this.getResult != null)
                    return this.getResult();

                if (this.getDefaultValue != null)
                    return this.getDefaultValue();

                throw new InvalidOperationException(string.Format("Value {0} is not valid.", this.input));
            }

            private SwitchBlock<TArg, TResult> Case(TArg value, Func<TResult> getResult)
            {
                if (getResult == null)
                    throw new ArgumentNullException("getResult");

                if (this.getResult == null && this.comparer.Equals(this.input, value))
                    this.getResult = getResult;

                return this;
            }
        }

        /// <summary>
        /// Represents a switch expression.
        /// </summary>
        /// <typeparam name="TResult">The type of the result.</typeparam>
        [Undisclosed]
        public interface ISwitch<TResult>
        {
            /// <summary>
            /// Ends the expression and returns the result.
            /// </summary>
            /// <returns>The result of the expression.</returns>
            /// <exception cref="InvalidOperationException">A matching case was not found and there was no default specified.</exception>
            TResult End();
        }

        /// <summary>
        /// Represents a switch expression without a default value.
        /// </summary>
        /// <typeparam name="TArg">The type of the input argument.</typeparam>
        /// <typeparam name="TResult">The type of the result.</typeparam>
        [Undisclosed]
        public interface IDefaultSwitch<TArg, TResult> : ISwitch<TResult>
        {
            /// <summary>
            /// Specifies the result of a specified input value.
            /// </summary>
            /// <param name="value">The input value.</param>
            /// <param name="result">The result to return for the specified input value.</param>
            /// <returns>The switch expression.</returns>
            IDefaultSwitch<TArg, TResult> Case(TArg value, TResult result);
            /// <summary>
            /// Specifies the result of a specified input value.
            /// </summary>
            /// <param name="value">The input value.</param>
            /// <param name="getResult">The delegate which returns a result for the specified input value.</param>
            /// <returns>The switch expression.</returns>
            /// <exception cref="ArgumentNullException"><paramref name="getResult"/> is null.</exception>
            IDefaultSwitch<TArg, TResult> Case(TArg value, Func<TResult> getResult);
            /// <summary>
            /// Specifies the result to return when none of the input values match.
            /// </summary>
            /// <param name="result">The result to return.</param>
            /// <returns>The switch expression.</returns>
            ISwitch<TArg, TResult> Default(TResult result);
            /// <summary>
            /// Specifies the result to return when none of the input values match.
            /// </summary>
            /// <param name="getResult">The delegate which returns a result.</param>
            /// <returns>The switch expression.</returns>
            /// <exception cref="ArgumentNullException"><paramref name="getResult"/> is null.</exception>
            ISwitch<TArg, TResult> Default(Func<TResult> getResult);
        }

        /// <summary>
        /// Represents a switch expression that has a default value.
        /// </summary>
        /// <typeparam name="TArg">The type of the input argument.</typeparam>
        /// <typeparam name="TResult">The type of the result.</typeparam>
        [Undisclosed]
        public interface ISwitch<TArg, TResult> : ISwitch<TResult>
        {
            /// <summary>
            /// Specifies the result of a specified input value.
            /// </summary>
            /// <param name="value">The input value.</param>
            /// <param name="result">The result to return for the specified input value.</param>
            /// <returns>The switch expression.</returns>
            ISwitch<TArg, TResult> Case(TArg value, TResult result);
            /// <summary>
            /// Specifies the result of a specified input value.
            /// </summary>
            /// <param name="value">The input value.</param>
            /// <param name="getResult">The delegate which returns a result for the specified input value.</param>
            /// <returns>The switch expression.</returns>
            /// <exception cref="ArgumentNullException"><paramref name="getResult"/> is null.</exception>
            ISwitch<TArg, TResult> Case(TArg value, Func<TResult> getResult);
        }

    }
}
